-- C_Struct.e
-- C Structure Manipulation Tools
-- Thanks to Pete Eberlein for code

include machine.e
include msgbox.e

-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- MATH

-----------------------------------------------------------------------------
global function or_all(sequence s)
-- or together all elements of a sequence
    atom result
    
    result = 0
    for i = 1 to length(s) do
        result = or_bits(result, s[i])
    end for
    return result
end function

-----------------------------------------------------------------------------
global function lo_word( atom a )
    -- return low word
    return and_bits( a, #FFFF )
end function

-----------------------------------------------------------------------------
global function hi_word( atom a )
    -- return high word
    return floor(a/power(2,16))
end function

-----------------------------------------------------------------------------
global function shortInt( integer i )

    -- converts numbers to short ints

    -- this needs the more exact number some time...
    if i < floor(65524/2) then
        return i
    else
        return i - 65534
    end if
    
end function


-----------------------------------------------------------------------------
global function abs( atom a )
    if a < 0 then
        return -a
    else
        return a
    end if
end function


-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- FILE I/O
-----------------------------------------------------------------------------
function readBytes( integer hFile, integer size )

    -- read bytes from file

    sequence s

    s = repeat( 0, size )    
    for i = 1 to size do   
        s[i] = getc( hFile )
    end for
    return s
    
end function


-----------------------------------------------------------------------------
function readLong( integer hFile )                       

    -- read a Long from the file

    return bytes_to_int( readBytes( hFile, 4 ) )
    
end function

    



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- DATA STRUCTURES

-----------------------------------------------------------------------------
sequence strings        -- stores addresses of strings used in structures
strings = {}
-----------------------------------------------------------------------------
function allot_string( sequence s )    

    -- allocate a string for a structure
    integer ok
    atom at
    
    -- place string in memory
    at = allocate_string( s )
    if at = 0 then
        -- out of string space
        ok = message_box( "Unable to allocate string space.", 
                            "Win32Lib Error", MB_ICONHAND+MB_TASKMODAL )
    
    
        -- abort
        abort(0)
        
    end if

    -- save location for garbage collection    
    strings = append( strings, at )
    
    -- return the address
    return at
    
end function

-----------------------------------------------------------------------------
global procedure free_strings()              

    -- free allocated strings
    
    -- for each allocated string...
    for i = 1 to length( strings ) do
        -- ...free string
        free( strings[i] )
    end for                     
    
    -- clear the address list
    strings = {}
    
end procedure


-----------------------------------------------------------------------------
global constant
    Byte    = -1,
    Word    = -2,
    Integer = Word,
    Long    = -3,
    Dword   = Long,
    Lpsz   = -4
                                   



integer allottedSize
allottedSize = 0
-----------------------------------------------------------------------------
global function allot( integer i )

    -- returns allotted size, and adds new field
    -- used to build a data structure
    
    integer soFar, diff, size

    -- save position    
    soFar = allottedSize

    -- if not a pre-defined type, make sure it gets a word boundary
    if i > 0 then

        diff = remainder( soFar, 4 )
        if diff then
            -- word align
            soFar = soFar + 4 - diff
        end if

        -- size is actual size
        size = i

    elsif i = Byte then
        size = 1
        
    elsif i = Word then
        size = 2
        
    elsif i = Long then
        size = 4

    elsif i = Lpsz then
        size = 4        
        
    end if

    -- allot space
    allottedSize = allottedSize + size

    -- return size and data type
    return { soFar, i }
    
end function



-----------------------------------------------------------------------------
global function allotted_size()

    -- returns allotted size, and clears size

    integer soFar
    
    soFar = allottedSize

    allottedSize = 0
    
    return soFar
    
end function
    

-----------------------------------------------------------------------------
global procedure store( atom struct, sequence s, object o )

    -- fetch the data based on the type
    integer offset, size
    
    -- rest data type
    offset  = s[1]
    size    = s[2]
    
    -- read, based on size    
    if      size = Byte then    
        -- poke a byte
        poke( struct+offset, o )
                
    elsif   size = Word then
        -- poke a word
            s = int_to_bytes( o )
            poke( struct+offset, s[1..2] )
                
    elsif   size = Long then    
        -- poke a long
        poke4( struct+offset, o )

    elsif   size = Lpsz then
        -- if atom, treat as long        
        if atom( o ) then
            poke4( struct+offset, o )
        else
            -- poke the address of the allotted string
            poke4( struct+offset, allot_string( o ) )
        end if
        
    else
        -- poke a string                        
        poke( struct+offset, o )
    end if

end procedure


-----------------------------------------------------------------------------
global function peek_string(atom a)

    -- read a lpsz string from address a
    -- Credit: Pete Eberlein
    
    integer i
    sequence s
    s = ""
    if a then
        i = peek(a)        
        while i do
            s = append(s, i)
            a = a + 1
            i = peek(a)
        end while
    end if
    return s
end function


-----------------------------------------------------------------------------
global function fetch( atom struct, sequence s )

    -- fetch the data based on the type
    integer size
    atom at
    
    -- address is struct + offset
    at = struct + s[1]
    
    -- get data size
    size    = s[2]    
    
    -- read, based on size    
    if      size = Byte then    
        -- return byte
        return peek( at )
        
    elsif   size = Word then    
        -- return word           
        return bytes_to_int( peek({at, 2}) & {0, 0} )
        
    elsif   size = Long then    
        -- return long
        return peek4s( at )
        
    elsif   size = Lpsz then            
        -- get the pointer
        at = peek4u(at)     
        
        -- return the string
        return peek_string( at )

    else
        -- return the string                        
        return( peek( {at, size} ) )
        
    end if

end function


-----------------------------------------------------------------------------
global function allocate_struct( integer size )

        -- allocate space for a structure
        -- and initialize to zero
        
        atom at
        
        at  = allocate( size )
        mem_set( at, 0, size )
        
        return at
        
end function


-----------------------------------------------------------------------------
global function address( atom addr, object o )
    -- return address in structure  
    if atom( o ) then
        return addr + o
    else
        return addr + o[1]
    end if
end function


-----------------------------------------------------------------------------
function loadStructFromFile( integer hFile, atom into, integer bytes )

    -- load structure on file directly into a memory structure
    sequence s

    -- the sequence will hold the data temporarily
    s = repeat( 0, bytes )
    
    -- for each byte in the structure
    for i = 1 to bytes do
        -- get a byte from the file
        s[i] = getc( hFile )
    end for
    
    -- look for a -1
    if find( -1, s ) then
        -- end of file, return failure
        return 0
    end if       

    -- load structure    
    poke( into, s )
    
    -- return success
    return 1

end function

